# -*- coding: utf-8 -*-
from Acquisition import aq_base
from plone.app.caching.interfaces import IPloneCacheSettings
from plone.registry.interfaces import IRegistry
from Products.CMFCore.interfaces import IDynamicType
from Products.CMFDynamicViewFTI.interfaces import IBrowserDefault
from zope.component import queryUtility


def isPurged(object):
    """Determine if object is of a content type that should be purged.

    This inspects ``purgedContentTypes`` in the registry.
    """

    registry = queryUtility(IRegistry)
    if registry is None:
        return False

    settings = registry.forInterface(IPloneCacheSettings, check=False)
    if not settings.purgedContentTypes:
        return False

    portal_type = getattr(aq_base(object), 'portal_type', None)
    if portal_type is None:
        return False

    return (portal_type in settings.purgedContentTypes)


def stripLeadingCharacters(name):
    """Strip off leading / and/or @@
    """

    if name and name[0] == '/':
        name = name[1:]
    if name and name.startswith('@@'):
        name = name[2:]

    return name


def getObjectDefaultView(context):
    """Get the id of an object's default view
    """

    # courtesy of Producs.CacheSetup

    browserDefault = IBrowserDefault(context, None)

    if browserDefault is not None:
        try:
            return stripLeadingCharacters(browserDefault.defaultView())
        except AttributeError:
            # Might happen if FTI didn't migrate yet.
            pass

    if not IDynamicType.providedBy(context):
        return None

    fti = context.getTypeInfo()
    try:
        # XXX: This isn't quite right since it assumes the action starts
        # with ${object_url}
        action = fti.getActionInfo('object/view')['url'].split('/')[-1]
    except ValueError:
        # If the action doesn't exist, stop
        return None

    # Try resolving method aliases because we need a real template_id here
    if action:
        action = fti.queryMethodID(action, default=action, context=context)
    else:
        action = fti.queryMethodID('(Default)', default=action,
                                   context=context)

    return stripLeadingCharacters(action)



# -*- coding: utf-8 -*-
from plone.app.caching.interfaces import IPloneCacheSettings
from plone.app.caching.testing import getToken
from plone.app.caching.testing import PLONE_APP_CACHING_FUNCTIONAL_TESTING
from plone.app.testing import applyProfile
from plone.app.testing import setRoles
from plone.app.testing import TEST_USER_ID
from plone.app.testing import TEST_USER_NAME
from plone.app.testing import TEST_USER_PASSWORD
from plone.app.textfield.value import RichTextValue
from plone.cachepurging.interfaces import ICachePurgingSettings
from plone.cachepurging.interfaces import IPurger
from plone.caching.interfaces import ICacheSettings
from plone.namedfile.file import NamedImage
from plone.registry.interfaces import IRegistry
from plone.testing.z2 import Browser
from zope.component import getUtility
from zope.globalrequest import setRequest

import OFS.Image
import pkg_resources
import unittest2 as unittest


TEST_IMAGE = pkg_resources.resource_filename(
    'plone.app.caching.tests', 'test.gif')
TEST_FILE = pkg_resources.resource_filename(
    'plone.app.caching.tests', 'test.gif')


class TestOperations(unittest.TestCase):
    """This test aims to exercise some generic caching operations in a semi-
    realistic scenario.

    The caching operations defined by GS profiles are grouped elsewhere.
    See `test_profile_without_caching_proxy.py` and
    `test_profile_with_caching_proxy.py`

    Note: Changes made using the API, accessing objects directly, are done
    with the Manager role set. Interactions are tested using the testbrowser.
    Unless explicitly logged in (e.g. by adding a HTTP Basic Authorization
    header), this accesses Plone as an anonymous user.

    The usual pattern is:

    * Configure caching settings
    * Set up test content
    * Create a new testbrowser
    * Set any request headers
    * Access content
    * Inspect response headers and body
    * Repeat as necessary

    To test purging, check the self.purger._sync and self.purger._async lists.
    """

    layer = PLONE_APP_CACHING_FUNCTIONAL_TESTING

    def setUp(self):
        self.app = self.layer['app']
        self.portal = self.layer['portal']

        setRequest(self.portal.REQUEST)

        applyProfile(self.portal, 'plone.app.caching:without-caching-proxy')

        self.registry = getUtility(IRegistry)

        self.cacheSettings = self.registry.forInterface(ICacheSettings)
        self.cachePurgingSettings = self.registry.forInterface(
            ICachePurgingSettings)
        self.ploneCacheSettings = self.registry.forInterface(
            IPloneCacheSettings)

        self.purger = getUtility(IPurger)
        self.purger.reset()

    def tearDown(self):
        setRequest(None)

    def test_disabled(self):
        self.cacheSettings.enabled = False

        setRoles(self.portal, TEST_USER_ID, ('Manager',))

        # Folder content
        self.portal.invokeFactory('Folder', 'f1')
        self.portal['f1'].title = u"Folder one"
        self.portal['f1'].description = u"Folder one description"
        self.portal['f1'].reindexObject()

        # Publish the folder
        self.portal.portal_workflow.doActionFor(self.portal['f1'], 'publish')

        # Non-folder content
        self.portal['f1'].invokeFactory('Document', 'd1')
        self.portal['f1']['d1'].title = u"Document one"
        self.portal['f1']['d1'].description = u"Document one description"
        self.portal['f1']['d1'].text = RichTextValue(
            u"<p>Body one</p>",
            'text/plain',
            'text/html'
        )
        self.portal['f1']['d1'].reindexObject()

        # Publish the document
        self.portal.portal_workflow.doActionFor(
            self.portal['f1']['d1'], 'publish')

        # Content image
        self.portal['f1'].invokeFactory('Image', 'i1')
        self.portal['f1']['i1'].title = u"Image one"
        self.portal['f1']['i1'].description = u"Image one description"
        self.portal['f1']['i1'].image = NamedImage(
            open(TEST_IMAGE, 'rb'), 'image/gif', u'test.gif')
        self.portal['f1']['i1'].reindexObject()

        # Content file
        self.portal['f1'].invokeFactory('File', 'f1')
        self.portal['f1']['f1'].title = u"File one"
        self.portal['f1']['f1'].description = u"File one description"
        self.portal['f1']['f1'].file = OFS.Image.File(
            'test.gif', 'test.gif', open(TEST_FILE, 'rb'))
        self.portal['f1']['f1'].reindexObject()

        # OFS image (custom folder)
        OFS.Image.manage_addImage(
            self.portal['portal_skins']['custom'],
            'test.gif',
            open(TEST_IMAGE, 'rb')
        )

        setRoles(self.portal, TEST_USER_ID, ('Member',))

        import transaction
        transaction.commit()
        browser = Browser(self.app)

        # Check that we can open all without errors and without cache headers
        browser.open(self.portal.absolute_url())
        self.assertFalse('Cache-Control' in browser.headers)

        browser.open(self.portal['f1'].absolute_url())
        self.assertTrue('Folder one description' in browser.contents)
        self.assertFalse('Cache-Control' in browser.headers)

        browser.open(self.portal['f1']['d1'].absolute_url())
        self.assertTrue('Document one description' in browser.contents)
        self.assertFalse('Cache-Control' in browser.headers)

        browser.open(self.portal['f1']['i1'].absolute_url())
        self.assertFalse('Cache-Control' in browser.headers)

        browser.open(self.portal['f1']['f1'].absolute_url())
        self.assertFalse('Cache-Control' in browser.headers)

        browser.open(
            self.portal.absolute_url() + '/portal_skins/custom/test.gif')
        self.assertFalse('Cache-Control' in browser.headers)

        browser.open(
            self.portal.absolute_url() + '/++resource++plone.app.caching.gif')
        # Set by resources themselves, but irrelevant to this test:
        # self.assertTrue('Cache-Control' in browser.headers)

    def test_auto_purge_content_types(self):

        setRoles(self.portal, TEST_USER_ID, ('Manager',))

        # Non-folder content
        self.portal.invokeFactory('Document', 'd1')
        self.portal['d1'].title = u"Document one"
        self.portal['d1'].description = u"Document one description"
        self.portal['d1'].text = RichTextValue(
            u"<p>Body one</p>",
            'text/plain',
            'text/html'
        )
        self.portal['d1'].reindexObject()

        setRoles(self.portal, TEST_USER_ID, ('Member',))

        # Purging disabled
        self.cachePurgingSettings.enabled = False
        self.cachePurgingSettings.cachingProxies = ()
        self.ploneCacheSettings.purgedContentTypes = ()

        editURL = '{0}/edit?_authenticator={1}'.format(
            self.portal['d1'].absolute_url(),
            getToken(TEST_USER_NAME)
        )

        import transaction
        transaction.commit()

        browser = Browser(self.app)
        browser.handleErrors = False
        browser.addHeader(
            'Authorization',
            'Basic {0}:{1}'.format(TEST_USER_NAME, TEST_USER_PASSWORD, ),
        )

        browser.open(editURL)

        browser.getControl(
            name='form.widgets.IDublinCore.title').value = u"Title 1"
        browser.getControl('Save').click()

        self.assertEqual([], self.purger._sync)
        self.assertEqual([], self.purger._async)

        # Enable purging, but not the content type
        self.cachePurgingSettings.enabled = True
        self.cachePurgingSettings.cachingProxies = ('http://localhost:1234',)
        self.ploneCacheSettings.purgedContentTypes = ()

        import transaction
        transaction.commit()

        browser.open(editURL)
        browser.getControl(
            name='form.widgets.IDublinCore.title').value = u"Title 2"
        browser.getControl('Save').click()

        self.assertEqual([], self.purger._sync)
        self.assertEqual([], self.purger._async)

        # Enable the content type, but disable purging
        self.cachePurgingSettings.enabled = False
        self.cachePurgingSettings.cachingProxies = ('http://localhost:1234',)
        self.ploneCacheSettings.purgedContentTypes = ()

        import transaction
        transaction.commit()

        browser.open(editURL)
        browser.getControl(
            name='form.widgets.IDublinCore.title').value = u"Title 3"
        browser.getControl('Save').click()

        self.assertEqual([], self.purger._sync)
        self.assertEqual([], self.purger._async)

        # Enable properly
        self.cachePurgingSettings.enabled = True
        self.cachePurgingSettings.cachingProxies = ('http://localhost:1234',)
        self.ploneCacheSettings.purgedContentTypes = ('Document',)

        import transaction
        transaction.commit()

        browser.open(editURL)
        browser.getControl(
            name='form.widgets.IDublinCore.title').value = u"Title 4"
        browser.getControl('Save').click()

        self.assertEqual([], self.purger._sync)
        self.assertEqual(set([
            'http://localhost:1234/plone/d1',
            'http://localhost:1234/plone/d1/document_view',
            'http://localhost:1234/plone/d1/',
            'http://localhost:1234/plone/d1/view']),
            set(self.purger._async))



# -*- coding: utf-8 -*-
from AccessControl.PermissionRole import rolesForPermissionOn
from plone.app.caching.interfaces import IETagValue
from plone.app.caching.interfaces import IPloneCacheSettings
from plone.app.caching.interfaces import IRAMCached
from plone.memoize.interfaces import ICacheChooser
from plone.registry.interfaces import IRegistry
from Products.CMFCore.interfaces import IContentish
from Products.CMFCore.interfaces import ISiteRoot
from thread import allocate_lock
from z3c.caching.interfaces import ILastModified
from zope.annotation.interfaces import IAnnotations
from zope.component import getUtility
from zope.component import queryMultiAdapter
from zope.component import queryUtility
from zope.interface import alsoProvides

import datetime
import dateutil.parser
import dateutil.tz
import logging
import re
import time
import wsgiref.handlers


PAGE_CACHE_KEY = 'plone.app.caching.operations.ramcache'
PAGE_CACHE_ANNOTATION_KEY = 'plone.app.caching.operations.ramcache.key'
ETAG_ANNOTATION_KEY = 'plone.app.caching.operations.etag'
LASTMODIFIED_ANNOTATION_KEY = 'plone.app.caching.operations.lastmodified'
_marker = object()

logger = logging.getLogger('plone.app.caching')

parseETagLock = allocate_lock()
# etagQuote = re.compile('(\s*\"([^\"]*)\"\s*,{0,1})')
# etagNoQuote = re.compile('(\s*([^,]*)\s*,{0,1})')

etagQuote = re.compile('(\s*(W\/)?\"([^\"]*)\"\s*,?)')
etagNoQuote = re.compile('(\s*(W\/)?([^,]*)\s*,?)')

#
# Operation helpers, used in the implementations of interceptResponse() and
# modifyResponse().
#
# These all take three parameters, published, request and response, as well
# as any additional keyword parameters required.
#


def setCacheHeaders(
    published,
    request,
    response,
    maxage=None,
    smaxage=None,
    etag=None,
    lastModified=None,
    vary=None
):
    """General purpose dispatcher to set various cache headers

    ``maxage`` is the cache timeout value in seconds
    ``smaxage`` is the proxy cache timeout value in seconds.
    ``lastModified`` is a datetime object for the last modified time
    ``etag`` is an etag string
    ``vary`` is a vary header string
    """

    if maxage:
        cacheInBrowserAndProxy(
            published,
            request,
            response,
            maxage,
            smaxage=smaxage,
            etag=etag,
            lastModified=lastModified,
            vary=vary
        )

    elif smaxage:
        cacheInProxy(
            published,
            request,
            response,
            smaxage,
            etag=etag,
            lastModified=lastModified,
            vary=vary
        )

    elif etag or lastModified:
        cacheInBrowser(
            published,
            request,
            response,
            etag=etag,
            lastModified=lastModified
        )

    else:
        doNotCache(published, request, response)


def doNotCache(published, request, response):
    """Set response headers to ensure that the response is not cached by
    web browsers or caching proxies.

    This is an IE-safe operation. Under certain conditions, IE chokes on
    ``no-cache`` and ``no-store`` cache-control tokens so instead we just
    expire immediately and disable validation.
    """

    if response.getHeader('Last-Modified'):
        del response.headers['last-modified']

    response.setHeader('Expires', formatDateTime(getExpiration(0)))
    response.setHeader('Cache-Control', 'max-age=0, must-revalidate, private')


def cacheInBrowser(published, request, response, etag=None, lastModified=None):
    """Set response headers to indicate that browsers should cache the
    response but expire immediately and revalidate the cache on every
    subsequent request.

    ``etag`` is a string value indicating an ETag to use.
    ``lastModified`` is a datetime object

    If neither etag nor lastModified is given then no validation is
    possible and this becomes equivalent to doNotCache()
    """

    if etag is not None:
        response.setHeader('ETag', '"{0}"'.format(etag), literal=1)

    if lastModified is not None:
        response.setHeader('Last-Modified', formatDateTime(lastModified))
    elif response.getHeader('Last-Modified'):
        del response.headers['last-modified']

    response.setHeader('Expires', formatDateTime(getExpiration(0)))
    response.setHeader('Cache-Control', 'max-age=0, must-revalidate, private')


def cacheInProxy(
    published,
    request,
    response,
    smaxage,
    etag=None,
    lastModified=None,
    vary=None
):
    """Set headers to cache the response in a caching proxy.

    ``smaxage`` is the timeout value in seconds.
    ``lastModified`` is a datetime object for the last modified time
    ``etag`` is an etag string
    ``vary`` is a vary header string
    """

    if lastModified is not None:
        response.setHeader(
            'Last-Modified',
            formatDateTime(lastModified)
        )
    elif response.getHeader('Last-Modified'):
        del response.headers['last-modified']

    if etag is not None:
        response.setHeader('ETag', '"{0}"'.format(etag), literal=1)

    if vary is not None:
        response.setHeader('Vary', vary)

    response.setHeader('Expires', formatDateTime(getExpiration(0)))
    response.setHeader(
        'Cache-Control',
        'max-age=0, s-maxage={0}, must-revalidate'.format(smaxage),
    )


def cacheInBrowserAndProxy(
    published,
    request,
    response,
    maxage,
    smaxage=None,
    etag=None,
    lastModified=None,
    vary=None
):
    """Set headers to cache the response in the browser and caching proxy if
    applicable.

    ``maxage`` is the timeout value in seconds
    ``smaxage`` is the proxy timeout value in seconds
    ``lastModified`` is a datetime object for the last modified time
    ``etag`` is an etag string
    ``vary`` is a vary header string
    """

    if lastModified is not None:
        response.setHeader('Last-Modified', formatDateTime(lastModified))
    elif response.getHeader('Last-Modified'):
        del response.headers['last-modified']

    if etag is not None:
        response.setHeader('ETag', '"{0}"'.format(etag), literal=1)

    if vary is not None:
        response.setHeader('Vary', vary)

    response.setHeader('Expires', formatDateTime(getExpiration(maxage)))

    if smaxage is not None:
        maxage = '{0}, s-maxage={1}'.format(maxage, smaxage)

    # Substituting proxy-validate in place of must=revalidate here because of
    # Safari bug
    # https://bugs.webkit.org/show_bug.cgi?id=13128
    response.setHeader(
        'Cache-Control',
        'max-age={0}, proxy-revalidate, public'.format(maxage)
    )


def cacheInRAM(
    published,
    request,
    response,
    etag=None,
    lastModified=None,
    annotationsKey=PAGE_CACHE_ANNOTATION_KEY
):
    """Set a flag indicating that the response for the given request
    should be cached in RAM.

    This will signal to a transform chain step after the response has been
    generated to store the result in the RAM cache.

    To actually use the cached response, you can implement
    ``interceptResponse()`` in your caching operation to call
    ``fetchFromRAMCache()`` and then return the value of the
    ``cachedResponse()`` helper.

    ``etag`` is a string identifying the resource.

    ``annotationsKey`` is the key used by the transform to look up the
    caching key when storing the response in the cache. It should match that
    passed to ``storeResponseInRAMCache()``.
    """

    annotations = IAnnotations(request, None)
    if annotations is None:
        return None

    key = getRAMCacheKey(request, etag=etag, lastModified=lastModified)

    annotations[annotationsKey] = key
    alsoProvides(request, IRAMCached)


def cachedResponse(
    published,
    request,
    response,
    status,
    headers,
    body,
    gzip=False
):
    """Returned a cached page. Modifies the response (status and headers)
    and returns the cached body.

    ``status`` is the cached HTTP status
    ``headers`` is a dictionary of cached HTTP headers
    ``body`` is a cached response body
    ``gzip`` should be set to True if the response is to be gzipped
    """

    response.setStatus(status)

    for k, v in headers.items():
        if k.lower() == 'etag':
            response.setHeader(k, v, literal=1)
        else:
            response.setHeader(k, v)

    response.setHeader('X-RAMCache', PAGE_CACHE_KEY, literal=1)

    if not gzip:
        response.enableHTTPCompression(request, disable=True)
    else:
        response.enableHTTPCompression(request)

    return body


def notModified(published, request, response, etag=None, lastModified=None):
    """Return a ``304 NOT MODIFIED`` response. Modifies the response (status)
    and returns an empty body to indicate the request should be interrupted.

    ``etag`` is an ETag to set on the response
    ``lastModified`` is the last modified date to set on the response

    Both ``etag`` and ``lastModified`` are optional.
    """

    if etag is not None:
        response.setHeader('ETag', etag, literal=1)

    # Specs say that Last-Modified MUST NOT be included in a 304
    # and Cache-Control/Expires MUST NOT be included unless they
    # differ from the original response.  We'll delete all, including
    # Expires although technically it should be included.  This is
    # probably okay since in the original we only include Expires
    # along with a Cache-Control and HTTP/1.1 clients will always
    # use the later over any Expires header anyway.  HTTP/1.0 clients
    # never send conditional requests so they will never see this.
    #
    # http://www.w3.org/Protocols/rfc2616/rfc2616-sec10.html
    #
    if response.getHeader('Last-Modified'):
        del response.headers['last-modified']
    if response.getHeader('Expires'):
        del response.headers['expires']
    if response.getHeader('Cache-Control'):
        del response.headers['cache-control']

    response.setStatus(304)
    return u''


#
# Cache checks
#

def cacheStop(request, rulename):
    """Check for any cache stop variables in the request.
    """

    # Only cache GET requests
    if request.get('REQUEST_METHOD') != 'GET':
        return True

    # rss_search also uses the SearchableText variable
    # so we'll hard code an exception for now
    if rulename == 'plone.content.feed':
        return False

    registry = getUtility(IRegistry)
    if registry is None:
        return False

    ploneSettings = registry.forInterface(IPloneCacheSettings)
    variables = ploneSettings.cacheStopRequestVariables
    return set(variables) & set(request.keys())


def isModified(request, etag=None, lastModified=None):
    """Return True or False depending on whether the published resource has
    been modified.

    ``etag`` is the current etag, to be checked against the If-None-Match
    header.

    ``lastModified`` is the current last-modified datetime, to be checked
    against the If-Modified-Since header.
    """

    if not etag and not lastModified:
        return True

    ifModifiedSince = request.getHeader('If-Modified-Since', None)
    ifNoneMatch = request.getHeader('If-None-Match', None)

    if ifModifiedSince is None and ifNoneMatch is None:
        return True

    etagMatched = False

    # Check etags
    if ifNoneMatch and etag is not None:
        if not etag:
            return True

        clientETags = parseETags(ifNoneMatch)
        if not clientETags:
            return True

        # is the current etag in the list of client-side etags?
        if etag not in clientETags and '*' not in clientETags:
            return True

        etagMatched = True

    """
    If a site turns off etags after having them on, the pages previously
    served will return an If-None-Match header, but the site will not be
    configured for etags. In this case, force a refresh to load the
    latest headers. I interpret this as the spec rule that the
    etags do NOT match, and therefor we must not return a 304.
    """
    if ifNoneMatch and etag is None:
        return True

    # Check the modification date
    if ifModifiedSince and lastModified is not None:

        # Attempt to get a date

        ifModifiedSince = ifModifiedSince.split(';')[0]
        ifModifiedSince = parseDateTime(ifModifiedSince)

        if ifModifiedSince is None:
            return True

        # has content been modified since the if-modified-since time?
        try:
            # browser only knows the date to one second resolution
            delta_sec = datetime.timedelta(seconds=1)
            if (lastModified - ifModifiedSince) > delta_sec:
                return True
        except TypeError:
            logger.exception('Could not compare dates')

        # If we expected an ETag and the client didn't give us one, consider
        # that an error. This may be more conservative than the spec requires.
        if etag is not None and not etagMatched:
            return True

    # XXX Do we really want the default here to be false?
    return False


def visibleToRole(published, role, permission='View'):
    """Determine if the published object would be visible to the given
    role.

    ``role`` is a role name, e.g. ``Anonymous``.
    ``permission`` is the permission to check for.
    """
    return role in rolesForPermissionOn(permission, published)

#
# Basic helper functions
#


def getContext(published, marker=(IContentish, ISiteRoot,)):
    """Given a published object, attempt to look up a context

    ``published`` is the object that was published.
    ``marker`` is a marker interface to look for

    Returns an item providing ``marker`` or None, if it cannot be found.
    """

    if not isinstance(marker, (list, tuple,)):
        marker = (marker,)

    def checkType(context):
        for m in marker:
            if m.providedBy(context):
                return True
        return False

    while (
        published is not None and
        not checkType(published) and
        getattr(published, '__parent__')
    ):
        published = published.__parent__

    if not checkType(published):
        return None

    return published


def formatDateTime(dt):
    """Format a Python datetime object as an RFC1123 date.

    If the datetime object is timezone-naive, it is assumed to be local time.
    """

    # We have to pass local time to format_date_time()

    if dt.tzinfo is not None:
        dt = dt.astimezone(dateutil.tz.tzlocal())

    return wsgiref.handlers.format_date_time(time.mktime(dt.timetuple()))


def parseDateTime(str):
    """Return a Python datetime object from an an RFC1123 date.

    Returns a datetime object with a timezone. If no timezone is found in the
    input string, assume local time.
    """

    try:
        dt = dateutil.parser.parse(str)
    except ValueError:
        return None

    if not dt:
        return None

    if dt.tzinfo is None:
        dt = datetime.datetime(dt.year, dt.month, dt.day,
                               dt.hour, dt.minute, dt.second, dt.microsecond,
                               dateutil.tz.tzlocal())

    return dt


def getLastModifiedAnnotation(published, request, lastModified=True):
    """Try to get the last modified date from a request annotation if available,
    otherwise try to get it from published object
    """

    if not lastModified:
        return None

    annotations = IAnnotations(request, None)
    if annotations is not None:
        dt = annotations.get(LASTMODIFIED_ANNOTATION_KEY, _marker)
        if dt is not _marker:
            return dt

    dt = getLastModified(published, lastModified=lastModified)

    if annotations is not None:
        annotations[LASTMODIFIED_ANNOTATION_KEY] = dt

    return dt


def getLastModified(published, lastModified=True):
    """Get a last modified date or None.

    If an ``ILastModified`` adapter can be found, and returns a date that is
    not timezone aware, assume it is local time and add timezone.
    """

    if not lastModified:
        return None

    lastModified = ILastModified(published, None)
    if lastModified is None:
        return None

    dt = lastModified()
    if dt is None:
        return None

    if dt.tzinfo is None:
        dt = datetime.datetime(dt.year, dt.month, dt.day,
                               dt.hour, dt.minute, dt.second, dt.microsecond,
                               dateutil.tz.tzlocal())

    return dt


def getExpiration(maxage):
    """Get an expiration date as a datetime in the local timezone.

    ``maxage`` is the maximum age of the item, in seconds. If it is 0 or
    negative, return a date ten years in the past.
    """

    now = datetime.datetime.now()
    if maxage > 0:
        return now + datetime.timedelta(seconds=maxage)
    else:
        return now - datetime.timedelta(days=3650)


def getETagAnnotation(published, request, keys=(), extraTokens=()):
    """Try to get the ETag from a request annotation if available,
    otherwise try to get it from published object
    """

    if not keys and not extraTokens:
        return None

    annotations = IAnnotations(request, None)
    if annotations is not None:
        etag = annotations.get(ETAG_ANNOTATION_KEY, _marker)
        if etag is not _marker:
            return etag

    etag = getETag(published, request, keys=keys, extraTokens=extraTokens)

    if annotations is not None:
        annotations[ETAG_ANNOTATION_KEY] = etag

    return etag


def getETag(published, request, keys=(), extraTokens=()):
    """Calculate an ETag.

    ``keys`` is a list of types of items to include in the ETag. These must
    match named multi-adapters on (published, request) providing
    ``IETagValue``.

    ``extraTokens`` is a list of additional ETag tokens to include, verbatim
    as strings.

    All tokens will be concatenated into an ETag string, separated by pipes.
    """

    if not keys and not extraTokens:
        return None

    tokens = []
    noTokens = True
    for key in keys:
        component = queryMultiAdapter(
            (published, request), IETagValue, name=key)
        if component is None:
            logger.warning(
                'Could not find value adapter for ETag component %s', key)
            tokens.append('')
        else:
            value = component()
            if value is None:
                value = ''
            else:
                noTokens = False
            tokens.append(value)

    for token in extraTokens:
        noTokens = False
        tokens.append(token)

    if noTokens:
        return None

    etag = '|' + '|'.join(tokens)
    etag = etag.replace(',', ';')  # commas are bad in etags
    etag = etag.replace('"', "'")  # double quotes are bad in etags

    return etag


def parseETags(text, allowWeak=True, _result=None):
    """Parse a header value into a list of etags. Handles fishy quoting and
    other browser quirks.

    ``text`` is the header value to parse.
    ``allowWeak`` should be False if weak ETag values should not be returned
    ``_result`` is internal - don't set it.

    Returns a list of strings. For weak etags, the W/ prefix is removed.
    """

    result = _result

    if result is None:
        result = []

    if not text:
        return result

    # Lock, since regular expressions are not threadsafe
    parseETagLock.acquire()
    try:
        m = etagQuote.match(text)
        if m:
            # Match quoted etag (spec-observing client)
            length = len(m.group(1))
            value = (m.group(2) or '') + (m.group(3) or '')
        else:
            # Match non-quoted etag (lazy client)
            m = etagNoQuote.match(text)
            if m:
                length = len(m.group(1))
                value = (m.group(2) or '') + (m.group(3) or '')
            else:
                return result
    finally:
        parseETagLock.release()

    if value:
        if value.startswith('W/'):
            if allowWeak:
                result.append(value[2:])
        else:
            result.append(value)

    return parseETags(text[length:], allowWeak=allowWeak, _result=result)

#
# RAM cache management
#


def getRAMCache(globalKey=PAGE_CACHE_KEY):
    """Get a RAM cache instance for the given key. The return value is ``None``
    if no RAM cache can be found, or a mapping object supporting at least
    ``__getitem__()``, ``__setitem__()`` and ``get()`` that can be used to get
    or set cache values.

    ``key`` is the global cache key, which must be unique site-wide. Most
    commonly, this will be the operation dotted name.
    """

    chooser = queryUtility(ICacheChooser)
    if chooser is None:
        return None

    return chooser(globalKey)


def getRAMCacheKey(request, etag=None, lastModified=None):
    """Calculate the cache key for pages cached in RAM.

    ``etag`` is a unique etag string.

    ``lastModified`` is a datetime object giving the last=modified
     date for the resource.

    The cache key is a combination of the resource's URL, the etag,
    and the last-modified date. Both the etag and last=modified are
    optional but in most cases that are worth caching in RAM, the etag
    is needed to ensure the key changes when the resource view changes.
    """

    resourceKey = '{0}{1}?{2}'.format(
        request.get('SERVER_URL', ''),
        request.get('PATH_INFO', ''),
        request.get('QUERY_STRING', ''),
    )
    if etag:
        resourceKey = '|' + etag + '||' + resourceKey
    if lastModified:
        resourceKey = '|' + str(lastModified) + '||' + resourceKey
    return resourceKey


def storeResponseInRAMCache(
    request,
    response,
    result,
    globalKey=PAGE_CACHE_KEY,
    annotationsKey=PAGE_CACHE_ANNOTATION_KEY
):
    """Store the given response in the RAM cache.

    ``result`` should be the response body as a string.

    ``globalKey`` is the global cache key. This needs to be the same key
    as the one used to fetch the data.

    ``annotationsKey`` is the key in annotations on the request from which
    the (resource-identifying) caching key should be retrieved. The default
    is that used by the ``cacheInRAM()`` helper function.
    """

    annotations = IAnnotations(request, None)
    if annotations is None:
        return

    key = annotations.get(annotationsKey)
    if not key:
        return

    cache = getRAMCache(globalKey)
    if cache is None:
        return

    """
    Resource registries have no body. If we put them in the cache the content
    type headers will indicate length and the body will be '', causing the
    browser to just spin. Furthermore, I doubt we ever want to cache an empty
    result: it's an indication that something went wrong somewhere.

    This does mean that any resources will not be cached in ram. There is
    potentially another fix but I doubt long term it's ever the right thing to
    do.
    """
    if result == '':
        return

    status = response.getStatus()
    headers = dict(request.response.headers)
    gzipFlag = response.enableHTTPCompression(query=True)

    cache[key] = (status, headers, result, gzipFlag)


def fetchFromRAMCache(
    request,
    etag=None,
    lastModified=None,
    globalKey=PAGE_CACHE_KEY,
    default=None
):
    """Return a page cached in RAM, or None if it cannot be found.

    The return value is a tuple as stored by ``storeResponseInRAMCache()``.

    ``etag`` is an ETag for the content, and is usually used as a basis for
    the cache key.

    ``lastModified`` is the last modified date for the content, which can
    potentially be used instead of etag if sufficient to ensure freshness.
    Perhaps a rare occurance but it's here in case someone needs it.
    Do not use this to cache binary responses (like images and file downloads)
    as Zope already caches most of the payload of these.

    ``globalKey`` is the global cache key. This needs to be the same key
    as the one used to store the data, i.e. it must correspond to the one
    used when calling ``storeResponseInRAMCache()``.
    """

    cache = getRAMCache(globalKey)
    if cache is None:
        return None

    key = getRAMCacheKey(request, etag=etag, lastModified=lastModified)
    if key is None:
        return None

    return cache.get(key, default)



# -*- coding: utf-8 -*-
from cStringIO import StringIO
from plone.app.caching.interfaces import IPloneCacheSettings
from plone.app.caching.testing import PLONE_APP_CACHING_FUNCTIONAL_TESTING
from plone.app.caching.tests.test_utils import stable_now
from plone.app.testing import applyProfile
from plone.app.testing import setRoles
from plone.app.testing import TEST_USER_ID
from plone.app.testing import TEST_USER_NAME
from plone.app.testing import TEST_USER_PASSWORD
from plone.app.textfield.value import RichTextValue
from plone.cachepurging.interfaces import ICachePurgingSettings
from plone.caching.interfaces import ICacheSettings
from plone.registry.interfaces import IRegistry
from plone.testing.z2 import Browser
from Products.CMFCore.FSFile import FSFile
from Products.CMFCore.utils import getToolByName
from zope.component import getUtility
from zope.globalrequest import setRequest

import datetime
import dateutil.parser
import dateutil.tz
import os
import pkg_resources
import unittest2 as unittest


TEST_FILE = pkg_resources.resource_filename(
    'plone.app.caching.tests', 'test.gif')


def test_image():
    from plone.namedfile.file import NamedBlobImage
    filename = pkg_resources.resource_filename(
        'plone.app.caching.tests', 'test.gif')
    filename = os.path.join(os.path.dirname(__file__), u'test.gif')
    return NamedBlobImage(
        data=open(filename, 'r').read(),
        filename=filename
    )


def _normalize_etag(s):
    s = s.split('|')
    s.pop()  # remove time-based component
    return '|'.join(s)


class TestProfileWithoutCaching(unittest.TestCase):
    """This test aims to exercise the caching operations expected from the
    `without-caching-proxy` profile.

    Several of the operations are just duplicates of the ones for the
    `with-caching-proxy` profile but we still want to redo them in this
    context to ensure the profile has set the correct settings.
    """

    layer = PLONE_APP_CACHING_FUNCTIONAL_TESTING

    def setUp(self):
        self.app = self.layer['app']
        self.portal = self.layer['portal']

        test_css = FSFile('test.css', os.path.join(
            os.path.dirname(__file__), 'test.css'))
        self.portal.portal_skins.custom._setOb('test.css', test_css)

        setRequest(self.portal.REQUEST)

        applyProfile(self.portal, 'plone.app.caching:without-caching-proxy')

        self.registry = getUtility(IRegistry)

        self.cacheSettings = self.registry.forInterface(ICacheSettings)
        self.cachePurgingSettings = self.registry.forInterface(
            ICachePurgingSettings)
        self.ploneCacheSettings = self.registry.forInterface(
            IPloneCacheSettings)

        self.cacheSettings.enabled = True

    def tearDown(self):
        setRequest(None)

    def test_composite_views(self):

        catalog = self.portal['portal_catalog']
        default_skin = self.portal['portal_skins'].default_skin

        # Add folder content
        setRoles(self.portal, TEST_USER_ID, ('Manager',))
        self.portal.invokeFactory('Folder', 'f1')
        self.portal['f1'].title = u"Folder one"
        self.portal['f1'].description = u"Folder one description"
        self.portal['f1'].reindexObject()

        # Add page content
        self.portal['f1'].invokeFactory('Document', 'd1')
        self.portal['f1']['d1'].title = u"Document one"
        self.portal['f1']['d1'].description = u"Document one description"
        testText = 'Testing... body one'
        self.portal['f1']['d1'].text = RichTextValue(
            testText,
            'text/plain',
            'text/html'
        )
        self.portal['f1']['d1'].reindexObject()

        # Publish the folder and page
        self.portal.portal_workflow.doActionFor(self.portal['f1'], 'publish')
        self.portal.portal_workflow.doActionFor(
            self.portal['f1']['d1'], 'publish')

        # Should we set up the etag components?
        # - set member?  No
        # - reset catalog counter?  Maybe
        # - set server language?
        # - turn on gzip?
        # - set skin?  Maybe
        # - leave status unlocked
        # - set the mod date on the resource registries?  Probably.

        import transaction
        transaction.commit()

        # Request the quthenticated folder
        now = stable_now()
        browser = Browser(self.app)
        browser.addHeader(
            'Authorization',
            'Basic {0}:{1}'.format(TEST_USER_NAME, TEST_USER_PASSWORD, ),
        )
        browser.open(self.portal['f1'].absolute_url())
        self.assertEqual('plone.content.folderView',
                         browser.headers['X-Cache-Rule'])
        self.assertEqual('plone.app.caching.weakCaching',
                         browser.headers['X-Cache-Operation'])
        # This should use cacheInBrowser
        self.assertEqual('max-age=0, must-revalidate, private',
                         browser.headers['Cache-Control'])
        tag = '"|test_user_1_|{0}|en|{1}|0|0'.format(
            catalog.getCounter(),
            default_skin,
        )
        self.assertEqual(tag, _normalize_etag(browser.headers['ETag']))
        self.assertTrue(now > dateutil.parser.parse(
            browser.headers['Expires']))

        # Set the copy/cut cookie and then request the folder view again
        browser.cookies.create('__cp', 'xxx')
        browser.open(self.portal['f1'].absolute_url())
        # The response should be the same as before except for the etag
        self.assertEqual('plone.content.folderView',
                         browser.headers['X-Cache-Rule'])
        self.assertEqual('plone.app.caching.weakCaching',
                         browser.headers['X-Cache-Operation'])
        self.assertEqual('max-age=0, must-revalidate, private',
                         browser.headers['Cache-Control'])
        tag = '"|test_user_1_|{0}|en|{1}|0|1'.format(
            catalog.getCounter(),
            default_skin,
        )
        self.assertEqual(tag, _normalize_etag(browser.headers['ETag']))

        # Request the authenticated page
        now = stable_now()
        browser = Browser(self.app)
        browser.addHeader(
            'Authorization',
            'Basic {0}:{1}'.format(TEST_USER_NAME, TEST_USER_PASSWORD, ),
        )
        browser.open(self.portal['f1']['d1'].absolute_url())
        self.assertTrue(testText in browser.contents)
        self.assertEqual('plone.content.itemView',
                         browser.headers['X-Cache-Rule'])
        self.assertEqual('plone.app.caching.weakCaching',
                         browser.headers['X-Cache-Operation'])
        # This should use cacheInBrowser
        self.assertEqual('max-age=0, must-revalidate, private',
                         browser.headers['Cache-Control'])
        tag = '"|test_user_1_|{0}|en|{1}|0'.format(
            catalog.getCounter(),
            default_skin,
        )
        self.assertEqual(tag, _normalize_etag(browser.headers['ETag']))
        self.assertTrue(now > dateutil.parser.parse(
            browser.headers['Expires']))

        # Request the authenticated page again -- to test RAM cache.
        browser = Browser(self.app)
        browser.addHeader(
            'Authorization',
            'Basic {0}:{1}'.format(TEST_USER_NAME, TEST_USER_PASSWORD, ),
        )
        browser.open(self.portal['f1']['d1'].absolute_url())
        self.assertEqual('plone.content.itemView',
                         browser.headers['X-Cache-Rule'])
        self.assertEqual('plone.app.caching.weakCaching',
                         browser.headers['X-Cache-Operation'])
        # Authenticated should NOT be RAM cached
        self.assertEqual(None, browser.headers.get('X-RAMCache'))

        # Request the authenticated page again -- with an INM header to test
        # 304
        etag = browser.headers['ETag']
        browser = Browser(self.app)
        browser.raiseHttpErrors = False  # we really do want to see the 304
        browser.addHeader(
            'Authorization',
            'Basic {0}:{1}'.format(TEST_USER_NAME, TEST_USER_PASSWORD, ),
        )
        browser.addHeader('If-None-Match', etag)
        browser.open(self.portal['f1']['d1'].absolute_url())
        # This should be a 304 response
        self.assertEqual('304 Not Modified', browser.headers['Status'])
        self.assertEqual('', browser.contents)

        # Request the anonymous folder
        now = stable_now()
        browser = Browser(self.app)
        browser.open(self.portal['f1'].absolute_url())
        self.assertEqual('plone.content.folderView',
                         browser.headers['X-Cache-Rule'])
        self.assertEqual('plone.app.caching.weakCaching',
                         browser.headers['X-Cache-Operation'])
        # This should use cacheInBrowser
        self.assertEqual('max-age=0, must-revalidate, private',
                         browser.headers['Cache-Control'])
        tag = '"||{0}|en|{1}|0|0'.format(catalog.getCounter(), default_skin)
        self.assertEqual(tag, _normalize_etag(browser.headers['ETag']))
        self.assertTrue(now > dateutil.parser.parse(
            browser.headers['Expires']))

        # Request the anonymous page
        now = stable_now()
        browser = Browser(self.app)
        browser.open(self.portal['f1']['d1'].absolute_url())
        self.assertEqual('plone.content.itemView',
                         browser.headers['X-Cache-Rule'])
        self.assertEqual('plone.app.caching.weakCaching',
                         browser.headers['X-Cache-Operation'])
        self.assertTrue(testText in browser.contents)
        # This should use cacheInBrowser
        self.assertEqual('max-age=0, must-revalidate, private',
                         browser.headers['Cache-Control'])
        tag = '"||{0}|en|{1}|0'.format(catalog.getCounter(), default_skin)
        self.assertEqual(tag, _normalize_etag(browser.headers['ETag']))
        self.assertTrue(now > dateutil.parser.parse(
            browser.headers['Expires']))

        # Request the anonymous page again -- to test RAM cache.
        # Anonymous should be RAM cached
        now = stable_now()
        browser = Browser(self.app)
        browser.open(self.portal['f1']['d1'].absolute_url())
        self.assertEqual('plone.content.itemView',
                         browser.headers['X-Cache-Rule'])
        self.assertEqual('plone.app.caching.weakCaching',
                         browser.headers['X-Cache-Operation'])
        # This should come from RAM cache
        self.assertEqual('plone.app.caching.operations.ramcache',
                         browser.headers['X-RAMCache'])
        self.assertTrue(testText in browser.contents)
        self.assertEqual('max-age=0, must-revalidate, private',
                         browser.headers['Cache-Control'])
        tag = '"||{0}|en|{1}|0'.format(catalog.getCounter(), default_skin)
        self.assertEqual(tag, _normalize_etag(browser.headers['ETag']))
        self.assertTrue(now > dateutil.parser.parse(
            browser.headers['Expires']))

        # Request the anonymous page again -- with an INM header to test 304.
        etag = browser.headers['ETag']
        browser = Browser(self.app)
        browser.raiseHttpErrors = False
        browser.addHeader('If-None-Match', etag)
        browser.open(self.portal['f1']['d1'].absolute_url())
        self.assertEqual('plone.content.itemView',
                         browser.headers['X-Cache-Rule'])
        self.assertEqual('plone.app.caching.weakCaching',
                         browser.headers['X-Cache-Operation'])
        # This should be a 304 response
        self.assertEqual('304 Not Modified', browser.headers['Status'])
        self.assertEqual('', browser.contents)

        # Edit the page to update the etag
        testText2 = 'Testing... body two'
        self.portal['f1']['d1'].text = RichTextValue(
            testText2,
            'text/plain',
            'text/html'
        )
        self.portal['f1']['d1'].reindexObject()

        transaction.commit()

        # Request the anonymous page again -- to test expiration of 304 and
        # RAM.
        etag = browser.headers['ETag']
        browser = Browser(self.app)
        browser.addHeader('If-None-Match', etag)
        browser.open(self.portal['f1']['d1'].absolute_url())
        self.assertEqual('plone.content.itemView',
                         browser.headers['X-Cache-Rule'])
        self.assertEqual('plone.app.caching.weakCaching',
                         browser.headers['X-Cache-Operation'])
        # The etag has changed so we should get a fresh page.
        self.assertEqual(None, browser.headers.get('X-RAMCache'))
        self.assertEqual('200 Ok', browser.headers['Status'])

    def test_content_feeds(self):

        catalog = self.portal['portal_catalog']
        default_skin = self.portal['portal_skins'].default_skin

        # Enable syndication
        setRoles(self.portal, TEST_USER_ID, ('Manager',))
        self.syndication = getToolByName(self.portal, 'portal_syndication')
        self.syndication.editProperties(isAllowed=True)
        self.syndication.enableSyndication(self.portal)

        import transaction
        transaction.commit()

        # Request the rss feed
        now = stable_now()
        browser = Browser(self.app)
        browser.open(self.portal.absolute_url() + '/RSS')
        self.assertEqual('plone.content.feed', browser.headers['X-Cache-Rule'])
        self.assertEqual('plone.app.caching.weakCaching',
                         browser.headers['X-Cache-Operation'])
        # This should use cacheInBrowser
        self.assertEqual('max-age=0, must-revalidate, private',
                         browser.headers['Cache-Control'])
        tag = '"||{0}|en|{1}"'.format(catalog.getCounter(), default_skin)
        self.assertEqual(tag, browser.headers['ETag'])
        self.assertTrue(now > dateutil.parser.parse(
            browser.headers['Expires']))

        # Request the rss feed again -- to test RAM cache
        now = stable_now()
        rssText = browser.contents
        browser = Browser(self.app)
        browser.open(self.portal.absolute_url() + '/RSS')
        self.assertEqual('plone.content.feed', browser.headers['X-Cache-Rule'])
        self.assertEqual('plone.app.caching.weakCaching',
                         browser.headers['X-Cache-Operation'])
        # This should come from RAM cache
        self.assertEqual('plone.app.caching.operations.ramcache',
                         browser.headers['X-RAMCache'])
        self.assertEqual(rssText, browser.contents)
        self.assertEqual('max-age=0, must-revalidate, private',
                         browser.headers['Cache-Control'])
        tag = '"||{0}|en|{1}"'.format(catalog.getCounter(), default_skin)
        self.assertEqual(tag, browser.headers['ETag'])
        self.assertTrue(now > dateutil.parser.parse(
            browser.headers['Expires']))

        # Request the rss feed again -- with an INM header to test 304.
        etag = browser.headers['ETag']
        browser = Browser(self.app)
        browser.raiseHttpErrors = False
        browser.addHeader('If-None-Match', etag)
        browser.open(self.portal.absolute_url() + '/RSS')
        self.assertEqual('plone.content.feed', browser.headers['X-Cache-Rule'])
        self.assertEqual('plone.app.caching.weakCaching',
                         browser.headers['X-Cache-Operation'])
        # This should be a 304 response
        self.assertEqual('304 Not Modified', browser.headers['Status'])
        self.assertEqual('', browser.contents)

        # Request the authenticated rss feed
        now = stable_now()
        browser = Browser(self.app)
        browser.addHeader(
            'Authorization', 'Basic {0}:{1}'.format(
                TEST_USER_NAME,
                TEST_USER_PASSWORD,
            )
        )
        browser.open(self.portal.absolute_url() + '/RSS')
        self.assertEqual('plone.content.feed', browser.headers['X-Cache-Rule'])
        self.assertEqual('plone.app.caching.weakCaching',
                         browser.headers['X-Cache-Operation'])
        # This should use cacheInBrowser
        self.assertEqual('max-age=0, must-revalidate, private',
                         browser.headers['Cache-Control'])
        self.assertEqual('"|test_user_1_|{0}|en|{1}"'.format(
            catalog.getCounter(), default_skin), browser.headers['ETag'])
        self.assertTrue(now > dateutil.parser.parse(
            browser.headers['Expires']))

        # Request the authenticated rss feed again -- to test RAM cache
        browser = Browser(self.app)
        browser.addHeader(
            'Authorization', 'Basic {0}:{1}'.format(
                TEST_USER_NAME,
                TEST_USER_PASSWORD,
            )
        )
        browser.open(self.portal.absolute_url() + '/RSS')
        self.assertEqual('plone.content.feed', browser.headers['X-Cache-Rule'])
        self.assertEqual('plone.app.caching.weakCaching',
                         browser.headers['X-Cache-Operation'])
        # Authenticated should NOT be RAM cached
        self.assertEqual(None, browser.headers.get('X-RAMCache'))

    def test_content_files(self):

        # Add folder content
        setRoles(self.portal, TEST_USER_ID, ('Manager',))
        self.portal.invokeFactory('Folder', 'f1')
        self.portal['f1'].title = u"Folder one"
        self.portal['f1'].description = u"Folder one description"
        self.portal['f1'].reindexObject()

        # Add content image
        self.portal['f1'].invokeFactory('Image', 'i1')
        self.portal['f1']['i1'].title = u"Image one"
        self.portal['f1']['i1'].description = u"Image one description"
        self.portal['f1']['i1'].image = test_image()
        self.portal['f1']['i1'].reindexObject()

        # Publish the folder
        self.portal.portal_workflow.doActionFor(self.portal['f1'], 'publish')

        import transaction
        transaction.commit()

        # Request the image
        now = stable_now()
        browser = Browser(self.app)
        browser.open(self.portal['f1']['i1'].absolute_url())
        self.assertEqual('plone.content.file', browser.headers['X-Cache-Rule'])
        self.assertEqual('plone.app.caching.weakCaching',
                         browser.headers['X-Cache-Operation'])
        # This should use cacheInBrowser
        self.assertEqual('max-age=0, must-revalidate, private',
                         browser.headers['Cache-Control'])
        # remove this when the next line works
        self.assertIsNotNone(browser.headers.get('Last-Modified'))
        self.assertTrue(now > dateutil.parser.parse(
            browser.headers['Expires']))

        # Request the image again -- with an IMS header to test 304
        lastmodified = browser.headers['Last-Modified']
        browser = Browser(self.app)
        browser.raiseHttpErrors = False
        browser.addHeader('If-Modified-Since', lastmodified)
        browser.open(self.portal['f1']['i1'].absolute_url())
        self.assertEqual('plone.content.file', browser.headers['X-Cache-Rule'])
        self.assertEqual('plone.app.caching.weakCaching',
                         browser.headers['X-Cache-Operation'])
        # This should be a 304 response
        self.assertEqual('304 Not Modified', browser.headers['Status'])
        self.assertEqual('', browser.contents)

        # Request an image scale
        now = stable_now()
        browser = Browser(self.app)
        browser.open(self.portal['f1'][
                     'i1'].absolute_url() + '/@@images/image/preview')
        self.assertEqual('plone.content.file', browser.headers['X-Cache-Rule'])
        self.assertEqual('plone.app.caching.weakCaching',
                         browser.headers['X-Cache-Operation'])
        # This should use cacheInBrowser
        self.assertEqual('max-age=0, must-revalidate, private',
                         browser.headers['Cache-Control'])
        # remove this when the next line works
        self.assertIsNotNone(browser.headers.get('Last-Modified'))
        self.assertTrue(now > dateutil.parser.parse(
            browser.headers['Expires']))

    def test_resources(self):

        import transaction
        transaction.commit()

        # Request a skin image
        now = stable_now()
        browser = Browser(self.app)
        browser.open(self.portal.absolute_url() + '/rss.png')
        self.assertEqual('plone.resource', browser.headers['X-Cache-Rule'])
        self.assertEqual('plone.app.caching.strongCaching',
                         browser.headers['X-Cache-Operation'])
        # This should use cacheInBrowserAndProxy
        self.assertEqual('max-age=86400, proxy-revalidate, public',
                         browser.headers['Cache-Control'])
        # remove this when the next line works
        self.assertIsNotNone(browser.headers.get('Last-Modified'))
        timedelta = dateutil.parser.parse(browser.headers['Expires']) - now
        self.assertTrue(timedelta > datetime.timedelta(seconds=86390))

        # Request the skin image again -- with an IMS header to test 304
        lastmodified = browser.headers['Last-Modified']
        browser = Browser(self.app)
        browser.raiseHttpErrors = False
        browser.addHeader('If-Modified-Since', lastmodified)
        browser.open(self.portal.absolute_url() + '/rss.png')
        self.assertEqual('plone.resource', browser.headers['X-Cache-Rule'])
        self.assertEqual('plone.app.caching.strongCaching',
                         browser.headers['X-Cache-Operation'])
        # This should be a 304 response
        self.assertEqual('304 Not Modified', browser.headers['Status'])
        self.assertEqual('', browser.contents)

        # Request a large datafile (over 64K) to test files that use
        # the "response.write()" function to initiate a streamed response.
        # This is of type OFS.Image.File but it should also apply to
        # large OFS.Image.Image, large non-blog ATImages/ATFiles, and
        # large Resource Registry cooked files, which all use the same
        # method to initiate a streamed response.
        s = 'a' * (1 << 16) * 3
        self.portal.manage_addFile('bigfile', file=StringIO(
            s), content_type='application/octet-stream')

        import transaction
        transaction.commit()

        browser = Browser(self.app)
        browser.open(self.portal['bigfile'].absolute_url())
        self.assertEqual('plone.resource', browser.headers['X-Cache-Rule'])
        self.assertEqual('plone.app.caching.strongCaching',
                         browser.headers['X-Cache-Operation'])
        # This should use cacheInBrowserAndProxy
        self.assertEqual('max-age=86400, proxy-revalidate, public',
                         browser.headers['Cache-Control'])
        # remove this when the next line works
        self.assertIsNotNone(browser.headers.get('Last-Modified'))
        timedelta = dateutil.parser.parse(browser.headers['Expires']) - now
        self.assertTrue(timedelta > datetime.timedelta(seconds=86390))

    def test_stable_resources(self):
        # We don't actually have any non-RR stable resources yet
        # What is the best way to test this?
        # Maybe not important since the RR test exercises the same code?
        pass



# -*- coding: utf-8 -*-
from setuptools import find_packages
from setuptools import setup


version = '1.2.15'

setup(
    name='plone.app.caching',
    version=version,
    description='Plone UI and default rules for plone.caching/z3c.caching',
    long_description=(open('README.rst').read() + '\n' +
                      open('CHANGES.rst').read()),
    classifiers=[
        'Environment :: Web Environment',
        'Framework :: Plone',
        'Framework :: Plone :: 5.0',
        'Framework :: Plone :: 5.1',
        'Framework :: Zope2',
        'License :: OSI Approved :: GNU General Public License (GPL)',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2.7',
    ],
    keywords='plone caching',
    author='Plone Foundation',
    author_email='plone-developers@lists.sourceforge.net',
    url='https://pypi.python.org/pypi/plone.app.caching',
    license='GPL version 2',
    packages=find_packages(exclude=['ez_setup']),
    namespace_packages=['plone', 'plone.app'],
    include_package_data=True,
    zip_safe=False,
    install_requires=[
        'setuptools',
        'python-dateutil',
        'plone.caching',
        'plone.cachepurging',
        'plone.app.registry >= 1.0b5',
        'zope.browserresource',
        'zope.interface',
        'zope.component',
        'zope.publisher',
        'zope.pagetemplate',
        'plone.memoize',
        'plone.protect',
        'plone.registry >= 1.0b4',
        'Products.CMFDynamicViewFTI',
        'Products.GenericSetup',
        'Products.CMFCore',
        'Products.statusmessages',
        'Zope2',
        'Acquisition',
        'plone.app.z3cform',
        'z3c.form',
        'z3c.zcmlhook',
    ],
    extras_require={
        'test': [
            'plone.app.contenttypes[test]',
            'plone.app.testing',
        ]
    },
    entry_points="""
    [z3c.autoinclude.plugin]
    target = plone
    """,
)
